import pygame
from pygame.constants import *
from data.Exceptions import *
from Vector import Vector, angle_vector
from data.Bullet import Bullet, Hook
import Event

ROTATE_LEFT = K_LEFT
ROTATE_RIGHT = K_RIGHT

# Decorators for handling keypress repeat-rate functionality
# State is kept in attrs in self. For a func named foo,
# key state is self.key_foo, and toggle state is self.state_foo.
def toggle(starts=False):
    def toggle_decorator(func):
        def handler(self, start, name=func.__name__):
            # Actually handle keypress
            if start == True and getattr(self, 'key_%s'%name, False) == False:
                setattr(self, 'state_%s' % name,
                        not getattr(self, 'state_%s'%name, starts))
                func(self, getattr(self, 'state_%s'%name))
                setattr(self, 'key_%s' % name, True)
            elif start == False:
                setattr(self, 'key_%s' % name, False)
        return handler
    return toggle_decorator

def periodic(period, rapid=False):
    def period_decorator(func, period=period, rapid=rapid):
        def handler(self, start, name=func.__name__):
            if start == True:
                last = getattr(self, 'last_%s'%name, -period)
                if self.game.frames - last >= period:
                    setattr(self, 'last_%s'%name, self.game.frames)
                    func(self)
            if start == False and rapid:
                setattr(self, 'last_%s'%name, self.game.frames - period)

        return handler
    return period_decorator
class Input(object):
    def __init__(self, game):
        self.game = game
        self.keys_pressed = {}
        self.skipping = 0
        self.game.input_disabled = False
        self.hook = None

    def thread(self):
        while True:
            self.handle_events()
            yield None

    def disable(self):
        self.game.input_disabled = True

    def enable(self):
        self.game.input_disabled = False

    def handle_events(self):
        keys_pressed = self.keys_pressed
        game = self.game
        for e in pygame.event.get():
            if e.type == pygame.constants.QUIT:
                print "Quitting"
                raise KeyboardInterrupt
            if e.type == KEYDOWN:
                keys_pressed[e.key] = True
            if e.type == KEYUP:
                if keys_pressed.has_key(e.key): #Sometimes spurious events are sent
                    for g in game.goals:
                        g.handle_key(e.key)

                    if self.key_functions.has_key(e.key):
                        self.key_functions[e.key](self, False)
                    del keys_pressed[e.key]
        for key in keys_pressed:
            if self.key_functions.has_key(key):
                self.key_functions[key](self, True)
        if self.game.input_disabled: return
        player = game.player
        self.rotate(player, keys_pressed)

    def die(self, start):
        self.game.player.health = 0

    def accel(self, start):
        game = self.game
        player = game.player
        if game.input_disabled: return
        m = game.geometry.bodies[player].getMass().mass
        if start:
            accel = (1.0/8)*angle_vector(player.angle)*m
        else:
            accel = Vector([0, 0])
        game.add_force(player, accel)
        p = game.geometry.bodies[player].getPosition()[:2]
        # 8 is the size of the thrustsprite.. FIXME
        game.thrustpos = p-(player.size+8)*angle_vector(player.angle)

    def rotate(self, player, keys_pressed):
        rotvel = 0
        if keys_pressed.has_key(ROTATE_LEFT): rotvel += 4
        if keys_pressed.has_key(ROTATE_RIGHT): rotvel -= 4
        player.angle = (player.angle + rotvel) % 360

    def stop(self, start):
        if self.game.input_disabled: return
        self.game.geometry.set_velo(self.game.player,Vector([0, 0]))

    @periodic(10)
    def shoot_bullet(self):
        game = self.game
        if game.input_disabled: return
        player = game.player
        game.shoot(player, Bullet(player, killstr=player.killstr, bounces=0, strength=20))

    @toggle()
    def shoot_hook(self, start):
        if self.game.input_disabled: return
        game = self.game
        player = game.player
        if game.hooked(player):
            game.unhook(player)
        elif self.hook and (self.hook in game.geometry.objects()):
            game.delete_object(self.hook)
        else:
            self.hook = game.shoot(player,Hook(player))

    @toggle()
    def pause(self, start):
        self.game.pause(start)
        self.game.input_disabled = start

    def fast(self, start):
        if self.game.input_disabled: return
        velo = 60 * angle_vector(self.game.player.angle)
        print 'new velo = ', velo
        self.game.geometry.set_velo(self.game.player, velo)

    def screenshot(self, start):
        self.game.post(Event.ScreenshotEvent())

    def skip_level(self, start):
        raise SkipLevel

    def quit(self, start):
        raise KeyboardInterrupt

    key_functions = {K_UP: accel, K_t: stop, K_a: shoot_bullet, K_f: fast,
                     K_o:shoot_hook, K_s:shoot_hook, K_q: quit,
                     K_ESCAPE: skip_level, K_c: screenshot,
                     K_p: pause}



