# Level.py -- primitives for loading levels
# Classes:
# * Shape. This represents either a circle or a polygon in something's
#   geometry.
#   A shape can be placed, in which case it is also a data.Object.
#   Pos, velo, etc. can be specified in Shape constructor.
import math
from pygame.rect import Rect
import pygame
import os
import random
from data import Object, Characters, Constants
import data.Rope
from data.Exceptions import *
import Image as ImageModule
import Event as EventModule
import Goals
from Vector import listify, avg2d, range2d

THICKNESS=15
# The globals that will be available to the level.
levelglobs = {}
currentfile = None

class Pattern(object):
    def __init__(self, patternlist):
        self.lst = patternlist

    def thread(self, game, shape, object):
        repeat = True
        locdict = {'self':ShapeWrapper(shape)}
        while repeat:
            repeat = False
            for item in self.lst:
                #print 'doing', item, shape
                if item == Repeat: repeat = True; continue
                if isinstance(item, str): exec(item, levelglobs, locdict); continue
                if callable(item): item(); continue
                velo, t = item
                if velo:
                    game.geometry.set_velo(object, velo)
                for i in range(t):
                    yield None

        #print 'pattern over for', shape
        yield None
class Repeat(object):
    '''Dummy object to specify that a pattern repeats.'''
    pass

class Align(object):
    def __init__(self, conditions):
        self.conditions = conditions

    def compute_pos(self, object):
        pos = [0, 0]
        for cond in self.conditions:
            #print cond
            a = AlignCond(*cond)
            i, p = a.compute_pos(object)
            pos[i] = p
        #print pos
        return pos

class AlignCond(object):
    '''Class to handle aligning one object with another object.'''
    OPPOSITES = {'left': 'right', 'right':'left', 'top':'bottom', 'bottom':'top'}
    CRITDIM = {'left':0, 'right':0, 'top':1, 'bottom':1}
    def __init__(self, side, target, targside=None):
        self.side = side
        self.target = target
        if not targside:
            targside = self.OPPOSITES[self.side]
        self.targside = targside

    def compute_pos(self, object):
        '''Computes pos to put object in order to satisfy constraints.'''
        d = self.CRITDIM[self.side]
        r = object.rect()
        targrect = self.target.rect()
        targrect.move_ip(self.target.pos)
        #print 'aligning', r, 'with', targrect, "'s", self.side, '--', self.targside
        #print getattr(targrect, self.side)
        oldpos = r.topleft
        setattr(r, self.targside, getattr(targrect, self.side))
        newpos = r.topleft
        #print newpos[d], oldpos[d]
        return d, (newpos[d] - oldpos[d])

def wait(event, game, loopfunc, playable=True):
    game.input_disabled = not playable
    event.start(game)
    try:
        while not event.check(game):
            if playable and game.player.health <= 0:
                #print "dying"
                raise GameOverException
            loopfunc()
    finally:
        event.stop(game)
    return event.value(game)

# Level.py: functionality for loading levels
def load(filename, game, loopfunc):
    def wait_internal(event):
        wait(event, game, loopfunc)
        
    def place(shape):
        object = shape.place(game)
        if not isinstance(shape, Scenery):
            game.post(EventModule.ObjectRefreshEvent(shape._object))
        return object

    def populate_globals(dct):
        actionfuncs = {Goals.Event: wait_internal, Shape: place}
        whitelist = {'len':len, 'True':True, 'False': False,
                     '__builtins__': None, 'math':math,
                     'range':range,
                     'Repeat': Repeat, 'THICKNESS':THICKNESS,
                     'GameOver':GameOverException,
                     'map': map, 'randint':random.randint
                     }
        dct.update(whitelist)
        from data import AI
        dct.update({'characters':Characters, 'ai':AI})
        dct.update({'wait':wait_internal, 'place':place})
        dct.update({'game':game, 'loopfunc':None})  #Maybe useful?
        dct.update({'Rope':data.Rope})
        dct.update(make_actions(game))
        dct.update(Constants.__dict__)

        for name, obj in globals().items()+Goals.__dict__.items():
            if isinstance(obj, type):
                for cl, func in actionfuncs.iteritems():
                    if issubclass(obj, cl):
                        dct[name.lower()] = compose(func,obj)
                        dct[name] = obj
                        #print name

    global currentfile
    currentfile = filename
    game.currentfile = filename
    populate_globals(levelglobs) # Could memoize this
    execfile(filename, levelglobs)
    return True
    
def compose(f, g):
    '''Simple function compositor which has the added benefit of retaining
    definitions of f and g in a new closure.'''
    def fog(*args, **kwargs):
        return f(g(*args, **kwargs))
    return fog

whitelist = {'__builtins__': {}, 'True': True, 'False': False}
def metadata(file):
    name, ext = os.path.splitext(file)
    newfile = "%s-info%s" % (name, ext)
    d = dict(whitelist)
    execfile(newfile, d)
    for key in whitelist.keys():
        del d[key] # Take out the obvious stuff
    d['filename'] = file
    return d
#return {'title': 'Training part 5', 'subtitle': 'n00bs in space'}

# Shapes:
class Shape(object):
    '''Class for a shape as used by the user.

    For "smart placement", we accept the following forms:

    Shape(rel=walls, pos=(400, 400)) # for instance to place an object "in" a room
    Shape(rel=(left, walls, (0, 400))) # "to the left of walls, where our y=0 matches their y=400"

    se/lf.parts: label -> geomnum
    '''
    def __init__(self, pos=(0,0), velo=(0,0), accel=(0,0), label=0,
                 pattern=None, rel=None, on_destroy=None,
                 image=None, texture=None, bg=None, texturebg=None,
                 inanim=None, outanim=None,
                 corner=None, remains=None, corpse=None,
                 start=None, end=None, startcorner=None, endcorner=None,
                 invertcorner=None, corners_at=None,
                 **kwargs):
        self.kwargs = kwargs
        self.pos = pos
        self.rel = rel
        self.velo = velo
        self.accel = accel
        self.label = label
        self.parts = {}
        if not pattern:
            pattern = []
        self.pattern = pattern
        self.on_destroy = on_destroy
        self.pic('image', locals())
        self.pic('texture', locals())
        # corner stuff should probably go in Walls
        # maybe make the name of the texture available in self.texturename?
        self.pic('texturebg', locals())
        self.pic('inanim', locals())
        self.pic('remains', locals())
        self.pic('corpse', locals())
        self.outanim = outanim
        if corners_at:
            self.corners_at = corners_at
        else:
            self.corners_at = []
        l = locals()
        self.unpack(l, 'bg', ['bgdx', 'bgdy'], [0, 0])
        self.pic('bg', l)
        self.unpack(l, 'startcorner', ['startcornerrot'], [None])
        self.unpack(l, 'endcorner', ['endcornerrot'], [None])
        self.unpack(l, 'start', ['startrot'], [None])
        self.unpack(l, 'end', ['endrot'], [None])
        self.maybe_pic_from('corner', 'texture', '-corner', l)
        self.maybe_pic_from('start', 'texture', '-start', l)
        self.maybe_pic_from('end', 'texture', '-end', l)
        self.maybe_pic_from('startcorner', 'texture', '-startcorner', l)
        self.maybe_pic_from('endcorner', 'texture', '-endcorner', l)
        self.maybe_pic_from('invertcorner', 'texture', '-invertcorner', l)
        #self.pic('outanim', locals())
        self.kwargs['remains'] = remains
        self.kwargs['corpse'] = self.corpse

    def unpack(self, locals, attr, vars, defaults):
        try:
            if len(locals[attr]) == len(vars)+1:
                for i in range(len(vars)):
                    setattr(self, vars[i], locals[attr][i+1])
                locals[attr] = locals[attr][0]
            else:
                raise TypeError, "set defaults"
        except:
            for i in range(len(vars)):
                setattr(self, vars[i], defaults[i])

    def pic(self, attr, locals):
        if isinstance(locals[attr], str) and locals[attr]:
            file = locals[attr]
            setattr(self, attr+'name', file)
            setattr(self, attr, ImageModule.find_image(file, currentfile))
        else:
            setattr(self, attr, locals[attr])

    def maybe_pic_from(self, attr, oldattr, append, locals):
        if locals[attr]:
            #print "locals[attr]", locals, attr
            self.pic(attr, locals)
        elif locals[oldattr] and not locals[attr] == False:
            file = locals[oldattr] + append
            locals[attr] = file
            try:
                self.pic(attr, locals)
            except ValueError:
                setattr(self, attr, None)
        else:
            setattr(self, attr, None)

    def add_extra(self, part, extra):
        self._object.extras.setdefault(part, []).append(extra)

    def rect(self):
        c = self.center()
        s = self.size()
        r = Rect(0, 0, *s)
        r.center = c
        return r

    def _make_object(self, kwargs):
        '''Returns the object that should be added to the game if placed.'''
        return Object.Object(**kwargs)

    def _add_to(self, game, object):
        '''Function which adds specified shape to the geometry of object.'''
        pass
        
    def preplace(self, game, filename=None, object=None):
        if self.inanim: game.make_inanim(self); return
        self.place(game, filename, object)
    
    def place(self, game, filename=None, object=None):
        '''Add an object to a game, using various parameters in self to fill
        in defaults in Geometry.add_object.'''
        if not object: object = self._make_object(self.kwargs)
        self._object = object
        geom = game.geometry
        # Compute position using rel argument. Add pos to allow
        # placing relative to this.
        if self.rel:
            if isinstance(self.rel, ShapeWrapper):
                #print 'relative to', self.rel
                p = game.geometry.get_pos(self.rel._object)
                #print 'which is at', p
                self.pos = map(lambda x, y: x+y, self.pos, p)
                self.rel = None
            # rel = walls
            # rel = ('left', walls, (ours, theirs))
            else:
                if len(self.rel) > 1 and isinstance(self.rel[1], ShapeWrapper):
                    # rel is just one condition
                    self.rel = [self.rel]
                a = Align(self.rel)
                p = a.compute_pos(self)
                self.pos = map(lambda x, y: x+y, self.pos, p)
                self.rel = None
                
        #print 'creating new body at', object, self.pos
        geom.new_body(object, self.pos, self.velo, self.accel)
        game.patterns[object] = Pattern(self.pattern).thread(game, self, object)
        if self.on_destroy:
            object.on_destroy = Pattern(self.on_destroy).thread(game, self, object)
        self._add_to(game, object)
        v = self.value()
        object.outanim = self.outanim
        object.remains = self.remains
        try:
            return map(ShapeWrapper, v)
        except:
            return ShapeWrapper(v)

    def value(self):
        return self

    # Functions to allow derived classes to specify geometric properties
    def center(self):
        '''Center of this object'''
        return (0, 0)

    def size(self):
        '''Size of this object -- be sure to include negative coordinates!'''
        return (0, 0)

    def insidesize(self):
        return self.size()

    def disable(self):
        pass

    def enable(self):
        pass

class Scenery(Shape):
    body = None
    def __init__(self, **kwargs):
        super(Scenery, self).__init__(**kwargs)

    def place(self, game, filename=None, object=None):
        self.game = game # Some scenery needs to know this
        if self.body: pass # Create a body object, use patterns?..
        self._add_to(game)
        return self
    
# For setting attributes in shapes which affect their objects.
# Typical use might be to set a position. If you set a ShapeWrapper's
# position, it will change the object (if it has been placed) or the
# shape (if it has not), which will later change the shape.  This is
# necessary so that Objects can be given to Align with (the Align
# takes ShapeWrappers and uses the Shapes).  There might be more
# elegant ways to do this -- convert Shape.rect into some kind of
# Object.rect? How would Package get the size of its contents? -- but
# I'm kind of skeptical.
class ShapeWrapper(object):
    def __init__(self, shape):
        self.__dict__['shape'] = shape
        
    def __setattr__(self, attr, val):
        try:
            setattr(self.__dict__['shape']._object, attr, val)
        except:
            setattr(self.__dict__['shape'], attr, val)

    def __getattr__(self, attr):
        if attr == 'shape': return self.__dict__['shape']
        try:
            return getattr(self.__dict__['shape']._object, attr)
        except:
            return getattr(self.__dict__['shape'], attr)
            

class Polygon(Shape):
    def __init__(self, *args, **kwargs):
        super(Polygon, self).__init__(**kwargs)
        self.points = listify(args)

    def _add_to(self, game, object):
        print "Boned!"
        game.geometry.add_polygon(object, self.label, self.points)

    def center(self):
        mx, my = range2d(self.points)
        return mx/2, my/2

    def size(self):
        return range2d(self.points)
    
class Circle(Shape):
    def __init__(self, radius=0, center=None, density=1, **kwargs):
        super(Circle, self).__init__(**kwargs)
        self.radius = radius
        if not center:
            center = (0, 0)
        self.centerpos = center
        self.density = density

    def _add_to(self, game, object):
        game.geometry.add_circle(object, self.label, self.radius, self.density, self.centerpos, image=self.image)

    def center(self):
        return self.centerpos

    def size(self):
        return (2*self.radius,)*2

class Box(Shape):
    def __init__(self, lengths=(0, 0), offset=(0, 0), density=1, **kwargs):
        super(Box, self).__init__(**kwargs)
        self.lengths = lengths
        self.offset = offset
        self.density = density

    def _add_to(self, game, object):
        rect = pygame.Rect(0, 0, *self.lengths)
        rect.center = self.offset
        game.geometry.add_box(object, self.label, rect, self.density, image=self.image, texture=self.texture)

    def center(self):
        return self.offset

    def size(self):
        return (self.lengths)
    

class Player(Shape):
    PLAYERSIZE=16
    def __init__(self, pos=None, **kwargs):
        kwargs.setdefault('pos', pos)
        kwargs.setdefault('image', 'ship')
        kwargs.setdefault('corpse', 'ship-dead')
        kwargs.setdefault('outanim', 'ship-death')
        super(Player, self).__init__(**kwargs)

    def _make_object(self, kwargs):
        return Object.Player(self.PLAYERSIZE, **kwargs)

    def _add_to(self, game, object):
        game.player = object
        game.geometry.add_circle(object, 0, self.PLAYERSIZE, density=1, image=self.image)

    def center(self):
        return (0, 0)

    def size(self):
        return (self.PLAYERSIZE*2,)*2

class Enemy(Shape):
    SIGN = {'left': -1, 'right':1, 'top':-1, 'bottom':1}
    def __init__(self, ai=None, angle=0, shapes=[], size=8, bulletpos=None, bulletspeed = None, health=1, attackimg=None, **kwargs):
        super(Enemy, self).__init__(**kwargs)
        if not ai: ai = AI.Stationary()
        self.ai = ai
        self.shapes = shapes
        self.angle = angle
        self.bulletpos = bulletpos
        self.shapesize = size
        self.health = health
        self.bulletspeed = bulletspeed
        self.pic('attackimg', locals())

    def _make_object(self, kwargs):
        e = Object.Enemy(ai=self.ai, size=self.shapesize, bulletpos=self.bulletpos, health=self.health, angle=self.angle, bulletspeed=self.bulletspeed, **kwargs)
        e.attackimg = self.attackimg # FIXME: is this a huge hack?
        # problem is, how to get attackimg (Shape attr) where Game can access
        # it when the obj shoots? Put in AI? Object.Enemy attr images = {} ?
        return e

    def _add_to(self, game, object):
        for shape in self.shapes:
            shape._add_to(game, object)
        game.add_enemy(object)

    def center(self):
        return avg2d([s.center() for s in self.shapes])

    def size(self):
        return range2d([s.size() for s in self.shapes])

    def disable(self):
        self.offai = self._object.ai
        self._object.ai = data.AI.Stationary()

    def enable(self):
        self._object.ai = self.offai

class _wall(object):
    def __init__(self, lp, p):
        lx, ly = lp
        x, y = p
        # Vector math by hand, yuk!
        dx, dy = x - lx, y - ly
        l = math.sqrt(dx**2 + dy**2)
        if dy == 0:
            if dx < 0:
                r = pygame.Rect(lx, ly, dx, THICKNESS)
                texturerot = 2*90
            else:
                r = pygame.Rect(lx, ly, dx, -THICKNESS)
                texturerot = 0*90
        else:
            if dy < 0:
                r = pygame.Rect(lx, ly, -THICKNESS, dy)
                texturerot = 3*90
            else:
                r = pygame.Rect(lx, ly, THICKNESS, dy)
                texturerot = 1*90
        r.normalize()
        self.p1 = lp
        self.p2 = p
        self.v = (dx, dy)
        self.l = l
        self.rect = r
        self.rot = texturerot
    
    def rectrot(self):
        return self.rect, self.rot

    def decorate(self, game, part, image, rel, rot = None):
        startpos = {0: 'bottomleft', 90: 'topleft', 180: 'topright',
                    270: 'bottomright'}
        endpos = {0: 'bottomright', 90: 'bottomleft', 180: 'topleft',
                  270: 'topright'}
        if part == 'end':
            p = self.p2
            attr = endpos[self.rot]
        else:
            p = self.p1
            attr = startpos[self.rot]
        if rot == None:
            rot = self.rot
        return game.add_image(Constants.LAYERGEOMS, p, image, rot, attr,
                              rel=rel)

class Walls(Shape):
    '''Shape representing a bunch of rectangles, each a wall.

    Walls([(0, 0), (200, 0), (200, 200), (0, 200)])
    Walls([(0, 0), (200, 0), 'rightwall', (200, 200), (0, 200)])

    closed (bool): wall between first and last point? (Default: False)
    counter (bool): points given counter-clockwise? (Default: False)

    Strings between two points "label" a given wall.

    Assumes all walls are vertical or horizontal right now.'''
    def __init__(self, *points, **kwargs):
        self.closed = kwargs.pop('closed', False)
        self.counter = kwargs.pop('counter', False)
        kwargs.setdefault('static',True)
        kwargs.setdefault('health',None) # invulnerable by default
        super(Walls, self).__init__(**kwargs)
        points = listify(points)
        if self.counter:
            points.reverse()

        points = self.strip_labels(points)

        self.points = []
        for i, t in enumerate(points):
            if len(t) > 2:
                x, y = points[0:1]
                self.parts[points[2]] = len(points)+i
                self.points.append((x, y))
            else:
                self.points.append(t)

        self.generate_walls()

    def strip_labels(self, points):
        # Loop invariant: points[0:i] are all tuples.
        # If points[i] is a string, i-1 is the last point.
        # Wall 0 is between points 0 and 1, wall 1 is between 1 and 2..
        # So if we get a string between points i-1 and i, this string
        # refers to wall i-1.
        i = 0
        while i < len(points):
            if isinstance(points[i], str):
                self.parts[points[i]] = i-1
                del points[i]
                # points[i] is no longer a string.
            else:
                i += 1
        return points
                
    def generate_walls(self):
        self.walls = []
        lp = self.points[0]
        r = Rect(0, 0, 0, 0)
        points = self.points[1:]
        if self.closed: points.append(self.points[0])
        for i, p in enumerate(points):
            w = _wall(lp, p)
            self.walls.append(w)
            lp = p
            rect, rot = w.rectrot()
            r = r.union(rect)
            
        self._rect = r

        
    def _add_to(self, game, object):
        def place_wall(i):
            r, texturerot = self.walls[i].rectrot()
            rot = 0
            #rot = math.atan2(dy, dx)
            box = game.geometry.add_box(object, i, r, rotate=rot, texture=self.texture, image=self.image)
            box.rot = texturerot
            # FIXME: 9, 0 is just nonsense for now
            return 9, 0, box
        
        def add_tri(i, p1, p2, p3):
            return
            game.geometry.add_polygon(object, i+len(self.points), [p1, p2, p3])

        def get_corner(lastrot, rot):
            if corners.has_key((lastrot, rot)):
                attr, crot = corners[(lastrot, rot)]
                invert = False
            else:
                attr, crot = corners[(rot, lastrot)]
                crot = (-crot) % 360
                invert = True
            return attr, crot, invert

        placecorners = [] # list for corners which we will add later
        def add_corner(p, lastrot, rot, image=self.corner,
                       invertimage=self.invertcorner, crot=None):
            if not image: return
            givencrot = crot
            if lastrot != None and rot != lastrot:
                attr, crot, invert = get_corner(lastrot, rot)
                #print 'corner at', p, lastrot, rot, self.corner, crot, attr
                if isinstance(givencrot, int): crot = givencrot
                if invert:
                    if invertimage:
                        crot = -crot
                        image = invertimage
                    else:
                        print "can't invertimage"
                        return # Don't bother
                args = ((Constants.LAYERGEOMS, p, image, crot, attr), {'rel':self._object})
                placecorners.append(args)

        lp = self.points[0]
        lastrot = None
        corners = {(0, 90): ("bottomleft", 0), (90, 180): ("topleft", 90),
                   (180, 270): ("topright", 180), (270, 0): ("bottomright", 270)}
        if self.bg:
            game.add_image(Constants.LAYERBG,
                           (self.bgdx, self.bgdy),
                           self.bg, rel=self._object)
        if self.texturebg:
            #print "adding texture", self.texturebgname, self._object
            game.add_image(Constants.LAYERBG,
                           (0, 0),
                           self.texturebg, rel=self._object, texture=self.insidesize())
        oldwallp1 = None
        points = self.points[1:]
        if self.closed:
            points.append(self.points[0])
        for i, p in enumerate(points):
            oldwallp2, newwallp1, box = place_wall(i)
            rot = box.rot
            add_corner(lp, lastrot, rot)
            if lastrot == None:
                firstrot = rot
            # XXX: if we ever add geometry triangles to round out
            # corners, this is how we do it
            if oldwallp1:
                add_tri(i, oldwallp1, oldwallp2, lp)
            else:
                firstp2 = oldwallp2
            oldwallp1 = newwallp1
                
            lastrot = rot
            lp = p

        # Force adding of geom sprites before scenery
        game.post(EventModule.ObjectRefreshEvent(self._object))

        if self.closed:
            add_tri(i, newwallp1, firstp2, self.points[0])
            add_corner(lp, lastrot, firstrot)
        else:
            rect, rot = self.walls[0].rectrot()
            p = self.points[0]
            if self.start:
                if not isinstance(self.startrot, int): # 0 is OK
                    self.startrot = (rot+180)%360
                self.corners_at.append((0, 'start', ('start', self.startrot)))
                
            if self.startcorner:
                #print 'startcorner', self.startcornername
                if isinstance(self.startcorner, bool):
                    add_corner(p, (rot-90)%360, rot, crot=self.startcornerrot)
                else:
                    add_corner(p, (rot-90)%360, rot, self.startcorner,
                               crot=self.startcornerrot)
                    
            rect, rot = self.walls[-1].rectrot()
            if self.end:
                #print 'end'
                if not isinstance(self.endrot, int): # 0 is OK
                    self.endrot = rot
                # with n points, walls use points (0, 1), (1, 2), ..,(n-2, n-1)
                # there are n-1 walls, which are numbered 0..n-2
                self.corners_at.append((-1, 'end', ('end', self.endrot)))

            if self.endcorner:
                p = self.points[-1]
                if isinstance(self.endcorner, bool):
                    add_corner(p, rot, (rot+90)%360)
                else:
                    add_corner(p, rot, (rot+90)%360, self.endcorner)

        n = len(self.walls)
        for c in self.corners_at:
            try:
                wallnum, side, (style, rot) = c
            except:
                wallnum, side, style = c
                rot = None
            wallnum = wallnum % n
            if hasattr(self, style):
                image = getattr(self, style)
            else:
                image = ImageModule.make_strip(style, currentfile)
            e = self.walls[wallnum].decorate(game, side, image, self._object,
                                             rot)
            self.add_extra(wallnum, e)

        for c in placecorners:
            args, kwargs = c
            e = game.add_image(*args, **kwargs)
            #self.add_extra(args[1], e)

    def center(self):
        return self._rect.center
        mx, my = range2d(self.points)
        return mx/2, my/2

    def size(self):
        return self._rect.size
        mx, my = range2d(self.points)
        #return (mx, my)
        return (mx+2*THICKNESS, my+2*THICKNESS)

    def insidesize(self):
        mx, my = range2d(self.points)
        return (mx, my)
            
class Door(Scenery):
    '''A door in a set of walls -- something that can open and reveal a floor.'''
    def __init__(self, object, part, **kwargs):
        # Remember, object is a shapewrapper.
        super(Door, self).__init__(**kwargs)
        self.shape = object.shape
        self.obj = object.shape._object
        self.part = part
        self.is_open = False
        #self._object = None

    def _add_to(self, game):
        num = self.shape.parts[self.part]
        g = game.geometry.geoms[self.obj][num]
        if not self.texturebg:
            texturebg = self.shape.texturename + '-opendoor'
            self.pic('texturebg', locals())
        game.post(EventModule.GeomBGEvent(g, texture=self.texturebg))
        if self.texture:
            g.getGeom().texture = self.texture
            game.post(EventModule.ObjectRefreshEvent(self.obj))
            if self.start:
                if self.startrot:
                    startrot = self.startrot
                else:
                    startrot = self.shape.walls[num].rot
                e = self.shape.walls[num].decorate(game, 'start', self.start,
                                                   self.obj, startrot)
                self.shape.add_extra(num, e)
            if self.end:
                if self.endrot:
                    endrot = self.endrot
                else:
                    endrot = self.shape.walls[num].rot
                e = self.shape.walls[num].decorate(game, 'end', self.end,
                                               self.obj, endrot)
                self.shape.add_extra(num, e)

    def open(self):
        if not self.is_open:
            self.game.delete_part(self.obj, self.shape.parts[self.part])
            self.is_open = True

class Image(Scenery):
    def __init__(self, object, part=None, side=None, flipv=False, fliph=False, layer=None, offset=(0, 0), start=None, repeat=True, **kwargs):
        #print kwargs['image']
        super(Image, self).__init__(**kwargs)
        self.image = self.image.flip(fliph, flipv)
        self.shape = object.shape
        self.obj = object.shape._object
        self.part = part
        self.side = side
        if layer == None:
            layer = Constants.LAYERBG
        self.layer = layer
        self.offset = offset
        self.start = start
        self.repeat = repeat

    def _add_to(self, game):
        #print self.image, self.obj
        dx, dy = 0, 0
        if self.part:
            label = self.shape.parts[self.part]
            g = game.geometry.geoms[self.obj][label]
            i = self.image
            # Hacks to allow Align to work properly
            # We need to create rects for both this object and the image
            # and set them on the correct objects.
            if g.shape == 'transform':
                g = g.getGeom()
                x, y = g.getPosition()[:2]
            if g.shape == 'circle':
                w = h = 2*g.radius
            if g.shape == 'box':
                w, h = g.size
            r = Rect(0, 0, w, h)
            r.center = x, y
            def grect():
                return r
            def irect():
                w, h = i.get_width(), i.get_height()
                return Rect(0, 0, w, h)
            g.rect = grect
            g.pos = (0, 0)
            #print g.rect(), g.pos
            i.rect = irect
            a = Align([(self.side, g)])
            dx, dy = a.compute_pos(i)
            #del g.rect
            #del g.pos
            del i.rect
        if self.offset:
            dx += self.offset[0]
            dy += self.offset[1]
        #print dx, dy
        game.post(EventModule.AddImageEvent(self.layer,
                                            (dx, dy), self.image,
                                            rel=self.obj, start=self.start,
                                            repeat=self.repeat))

class FX(Scenery):
    def __init__(self, part, type, **kwargs):
        self.part = part
        self.type = type
        self.kwargs = kwargs
    def _add_to(self, game):
        e = EventModule.FXEvent(self.part, self.type, self.kwargs)
        game.post(e)
        self.sprite = e.response

    def kill(self):
        self.sprite.kill()

class Chain(Scenery):
    def __init__(self, obj1, obj2, pos1, pos2, start=0, **kwargs):
        self.rot = kwargs.pop('rot', 0)
        super(Chain, self).__init__(**kwargs)
        
        self.obj1 = obj1.shape._object
        self.obj2 = obj2.shape._object
        self.pos1 = pos1
        self.pos2 = pos2
        self.start = start

    def _add_to(self, game):
        game.post(EventModule.AddChainEvent(self.obj1, self.obj2, self.pos1, self.pos2, self.texture, self.rot, start=self.start))

class Music(Scenery):
    def __init__(self, filename, loop=True):
        self.mfile = filename
        self.loop = loop

    def _add_to(self, game):
        game.play_music(self.mfile, self.loop)

class Package(Walls):
    PADDING=10
    def __init__(self, payload=None, **kwargs):
        kwargs.setdefault('closed', True) # can you imagine an open package?
        if kwargs.has_key('bg'):
            kwargs.setdefault('texturebg', kwargs.pop('bg'))
        self.payload = payload
        sx, sy = payload.size()
        self.s = (sx+2*self.PADDING, sy+2*self.PADDING)

        super(Package, self).__init__([(0, 0), 'top', (self.s[0], 0),
                                       'right', self.s, 'bottom',
                                       (0, self.s[1]), 'left',], **kwargs)
        
        payload.pattern[:0] = self.motion(self.pattern)        
        payload.pattern[:0] = ['self.disable()']
        if payload.velo != (0, 0):
            payload.pattern.append((payload.velo, 1))
        payload.pattern.append('self.enable()')
        
    def _add_to(self, game, object=None):
        super(Package, self)._add_to(game, object)
        sx, sy = self.payload.size()
        r = Rect(0, 0, sx, sy)
        #print r, r.center
        r.center = self.payload.center()
        #print r, r.move(*self.pos),
        r.move_ip(*self.pos)
        r.move_ip(*self.center())
        self.payload.pos = r.center
        self.payshape = self.payload.place(game) # ie. create a new object
        game.post(EventModule.ObjectRefreshEvent(self.payshape._object))

    def motion(self, pattern):
        '''Return only the motion parts of the pattern.'''
        return filter(lambda x: isinstance(x, tuple), pattern)

    def center(self):
        return self.s[0]/2, self.s[1]/2
    
    def size(self):
        return self.s

    def value(self):
        return self, self.payshape

import data
class EnemyBox(Package):
    def __init__(self, payload=None, target=None, targetcut=None, side=None,
                 pos=None, **kwargs):
        
        self.target = target
        SIGN = {'left': -1, 'right':1, 'top':-1, 'bottom':1}
        def vector(dir, x):
            v = {'left':(1, 0), 'right':(-1, 0), 'top':(0, 1), 'bottom':(0, -1)}
            return v[dir][0]*x, v[dir][1]*x
        # Amount to move Payload in.

        rel = (side, target)
        boxpos = [0, 0]
        boxpos[1-AlignCond.CRITDIM[side]] = pos[1-AlignCond.CRITDIM[side]]
        boxpos[AlignCond.CRITDIM[side]] = 500*SIGN[side]
        inamt = pos[AlignCond.CRITDIM[side]]
        intime = float(abs(inamt)+payload.size()[0]+20)/30
        #print 'constructing enemybox', inamt, intime
        if self.target.parts.has_key(targetcut):
            tcutact = 'delete(self.target, "%s")'%targetcut
        else:
            tcutact = '%s.open()'%targetcut
        pattern = [(vector(side, 5), 100), ((0, 0), 0),
                   'delete(self, "%s")'%AlignCond.OPPOSITES[side],
                   tcutact]
        payload.pattern[:0] = [(vector(side,intime),30), ((0, 0), 0)]
        #print payload.pattern

        super(EnemyBox, self).__init__(pos=boxpos, rel=rel, payload=payload, pattern=pattern, **kwargs)

class Switch(Shape):
    def __init__(self, shapes=[], action=[], **kwargs):
        kwargs.setdefault('static', True)
        super(Switch, self).__init__(**kwargs)
        self.shapes = shapes
        self.action = action

    def _make_object(self, kwargs):
        return Object.Switch(action=Pattern(self.action), **kwargs)
    
    def _add_to(self, game, object):
        # FIXME: overlap with Enemy. collapse into Compound class?
        for shape in self.shapes:
            shape._add_to(game, object)

        
def mirror(pattern, axis):
    axmult = [1, 1]
    axmult[axis] = -1
    newpat = []
    for item in pattern:
        try:
            v, t = item
            newv = (v[0]*axmult[0], v[1]*axmult[1])
            newpat.append((newv, t))
        except:
            newpat.append(item)
    return newpat

def make_actions(game):
    '''Creates the functions which need to access game when called.

    Returns a dict which can be used to update levelglobs.'''
    def delete(shape, label=None):
        #print 'deleting', shape, label
        if isinstance(label, str):
            label = shape.parts[label]
        if label != None:
            game.delete_part(shape._object, label)
        else:
            game.delete_object(shape._object, chrome=False)
        
    def move_to(shape, pos):
        # Should not always move the center of the object!
        # That would destroy important relative relationships between, say,
        # a package (where (0, 0) is upper left) and its payload
        # (where (0, 0) is its center).
        print "moving_to"
        game.geometry.set_pos(shape._object, pos)

    def move_by(shape, dist):
        p0 = game.geometry.get_pos(shape._object)
        newp = dist[0] + p0[0], dist[1] + p0[1]
        game.geometry.set_pos(shape._object, newp)
        for r in game.ropes.values():
            if r.obj == shape._object:
                r.move(dist)

    def start(shape, pattern):
        game.patterns[shape._object] = Pattern(pattern).thread(game, shape, shape._object)

    def align(shape, side, rel, pos=None):
        '''Aligns shape correctly, and returns offset that was necessary
        to align it correctly (to apply to other objects).'''
        a = Align([(side, rel)])
        #print 'aligning'
        apos = a.compute_pos(shape)
        newpos = apos[0] + pos[0], apos[1] + pos[1]
        oldpos = game.geometry.get_pos(shape._object)
        #print 'oldpos', oldpos, 'newpos', newpos
        dist = [newpos[0]-oldpos[0], newpos[1]-oldpos[1]]
        move_by(shape, dist)
        return dist

    def mirror_y(pattern):
        return mirror(pattern, 1)

    def mirror_x(pattern):
        return mirror(pattern, 0)

    def set_flag(flag, value):
        game.flags[flag] = value

    def get_flag(flag):
        return game.flags.get(flag, None)

    def stop(sound):
        game.post(EventModule.StopSoundEvent(sound))

    def play(sound, loop=None):
        event = EventModule.NewSoundEvent(sound, loop, False)
        game.post(event)
        return event.sound

    return {'delete':delete, 'move_to': move_to, 'move_by': move_by,
            'align': align, 'set_flag': set_flag, 'get_flag': get_flag,
            'start':start, 'mirror_y':mirror_y,
            'mirror_x':mirror_x, 'stop':stop, 'play':play}
